#!/usr/bin/python

#
##
### MoThEr. Mommas gonna make all of your nightmares come true.
##                                                        PF
#

"""
Where the main Mother Classes are defined:

 * DbMother
 * MotherManager
 * MotherBox
"""
            
from eccez import *
from speaker import Speaker, ERR_COL, INF_COL, OKI_COL
from dbda import Psygres


#
##
### Mother Symbols
##
#

#
## MoThEr: init flags.
#

MO_NOA    = 0     # No Action
MO_DEL    = 1     # Del Action
MO_UP     = 2     # Update Action
MO_SAVE   = 3     # Save Action
MO_LOAD   = 4     # Load Action
MO_ULOAD  = 5     # Unsafe Load Action

_MO_MAX_FLAG = MO_ULOAD

MO_BEFORE = 0
MO_AFTER  = 1

#
## MoThEr: SQL symbols.
#

SQL_DEFAULT = "MOTHER_SQL_DEFAULT"
SQL_NULL    =  None
SQL_TRUE    = "True"
SQL_FALSE   = "False"

_DEFAULT_MO_SESSION= "MoSession-"


#
##
### MoThEr Utils.
##
#

def init_mother(cfile, pg_conn= None, fnaming= None):

    from speaker import init_speaker
    from commons import MotherNaming
    from dbda import init_dbda

    if DbMother._mo_initialized:
        Speaker.log_warning("Mother already initialized.")
        return

    DbMother._mo_cfile= cfile
    
    d= {}
    execfile(cfile, d)
    DbMother._mo_dbstruct= d['MOTHER_MAP']

    init_speaker(d)
    init_dbda(d, pg_conn)

    fnaming= fnaming or MotherNaming
    DbMother.MotherNaming= staticmethod(fnaming)

    DbMother._mo_initialized= True
    DbMother.__init__= DbMother.__post_init__


def MotherSession(name= None, pg_conn= None):

    from dbda import PsygresPool, PsygresFly

    if not name:
        import random
        name= "%s%d" % (_DEFAULT_MO_SESSION, random.randint(1,10000))

    Speaker.log_info("Initializing session %s", INF_COL(name))

    return PsygresPool._pool_initialized \
        and PsygresPool.newSession(name) \
        or  PsygresFly(pg_conn, name)

#
##
### AVoid import, use execfile.
##
#

def MotherMap(*args):

    from mother.commons import somethingFromFile
    try:
        return somethingFromFile(DbMother._mo_dbstruct, *args)
    except:
        Speaker.log_int_raise(
            "Error loading Mother Map; file %s is "
            "invalid path or mothermapper.py has to be launched.", 
            ERR_COL(DbMother._mo_dbstruct))


#
## Use these...
#

def MotherMappedFields(): 
    return MotherMap('TABLE_FIELDS_DICT')[0]

def MotherMappedPkeys():
    return MotherMap('TABLE_PKEYS_DICT')[0]

def MotherMappedTables():
    return MotherMap('TABLE_FIELDS_DICT', 'TABLE_PKEYS_DICT')

def MotherMappedChildren():
    return MotherMap('TABLE_CHILDS_DICT')[0]

def MotherMappedRels():
    return MotherMap('TABLE_RELATIONS_DICT')[0]

def MotherMappedDeps():
    return MotherMap('TABLE_CHILDS_DICT', 'TABLE_RELATIONS_DICT')


#
## Internals..
#

_MO_FLAGS_STRS = [
            'MO_NOA',
            'MO_DEL',
            'MO_UP',
            'MO_SAVE',
            'MO_LOAD',
            'MO_ULOAD'
            ]
_MO_WHEN_STRS = [
            'Before',
            'After'
            ]

_MO_FUCK=" MoThEr fucker!"

_FALSE_VALUES= [SQL_NULL,SQL_DEFAULT]

#
##
### The Mother Class
##
#

class DbMother(Psygres):
    """ The Holy MoThEr 
        An abstract class which represents a db Table.
    """
    _mo_initialized= False
    _mo_cfile= None
    _mo_dbstruct= None
        
    # Mapping actions
    @staticmethod
    def _flag_actions(obj, flag):
        if flag == MO_UP: return obj.update
        if flag == MO_NOA: return lambda *args: None
        if flag == MO_DEL: return obj.delete
        if flag == MO_SAVE: return obj.insert
        if flag == MO_LOAD: return onj.load
        if flag == MO_ULOAD: return obj.uload 
        obj.log_int_raise(_MO_FUCK)

    @staticmethod
    def _trigger_actions(obj, flag):
        if flag == MO_UP: return obj._update
        if flag == MO_DEL: return obj._delete
        if flag == MO_SAVE: return obj._insert
        if flag == MO_LOAD: return obj._load
        obj.log_int_raise(_MO_FUCK)

    def _if_not_initialized(self):
        if not DbMother._mo_initialized:
            self.log_int_raise("Mother not Initialized yet.")

    def __init__(self, store= {}, flag= MO_NOA, oid= None, session= None):
        """ __init__(dict_values, flag, oid, session)

            Never set oid! Oid is for MotherBox only.
            With session you can use your own Session.
        """

        self._if_not_initialized()
        # This is an hack to bypass initialization control
        # after the first Mother initialization
        self.__post_init__(store, flag, oid, session)


    def __post_init__(self, store= {}, flag= MO_NOA, oid= None, session= None):
        """ __init__(dict_values, flag, oid, session)

            Never set oid! Oid is for MotherBox only.
            With session you can use your own Session.
        """

        # table_name is needed...
        if not hasattr(self, 'table_name'):
            self.log_int_raise("No table_name, %s", _MO_FUCK)

        # Useful var
        tbl= self.table_name

        # self.fields is a list with TABLE fields.
        # self.pkeys is a list with primary keys for the TABLE.
        # You can set these vars in the child class or let
        # MoThEr assign them dynamically.
        if not hasattr(self, "fields"):
            tfd= MotherMappedFields()
            self.fields = tfd[tbl]
            del tfd
           
        if not hasattr(self, "pkeys"):
            tfd= MotherMappedPkeys()
            self.pkeys = tfd[tbl]
            del tfd

        # Frozeset!
        self.fields= frozenset(self.fields)
        self.pkeys= frozenset(self.pkeys)

        # Dynamic vars
        self.oid=oid


        # Use Universe() if specified
        if session:
            session._import(self)
        self.session= session

        # Storing dict values with some controls depending on the 
        # initialization flag....
        self._initStore(store, flag)
        # ... and enjoy ;)

    def initTriggers(self):
        """ initTriggers() -> None

        Initialize triggers. Don't call it directly. Use
        addTrigger instead.\n"""

        self.triggers_map= {}
        t= self.triggers_map
            
        for flag in [MO_UP, MO_LOAD, MO_SAVE, MO_DEL]:
            t[flag]= {}
            for when in [MO_AFTER, MO_BEFORE]:
                t[flag][when]= []

        self._trigger_initialized= True

    def _initStore(self, store, flag):
        """ _initStore(store, flag) --> None

        Inizialization of self._store depending on flag.
        """

        # Controls: trivial
        setK= set(store.keys())
        accio= self._invalidFields(setK)
        if accio:
            self.log_int_raise("Invalid Fields %s for table |%s|.",\
                                ERR_COL(list(accio)), self.table_name)

        self._moved=set([])
        self._store=store.copy()

        # Very Bad Flag. CRASH!
        if flag<0 or flag>_MO_MAX_FLAG:
            self.log_int_raise("Mth: invalid Flag %s, %s",\
                        ERR_COL(_MO_FLAGS_STRS[flag]), _MO_FUCK)

        # Signal Mother that passed fields are
        # new fields (if not oid)
        if not self.oid:
            self._moved= set([ 
                k for k in setK 
                if k not in self.pkeys])

        f= self._flag_actions(self, flag)
        f()

#
##
### Internal Core
##
#
    def _triggers_are_initialized(self):
        return hasattr(self, 'triggers_map')

    def _unvaluedPKeys(self):
        """ _unvaluedPKeys() --> set

        Returns the set of pkeys not valu'ed.
        NOTE: a pkeys is considere NOT VALUED if his value is 
        DEFAULT or NULL.
        """

        s=self._store
        return set([\
            k for k in self.pkeys\
            if not s.has_key(k)\
            or s[k] in _FALSE_VALUES])

    def _noPKeyFields(self):
        """ _noPKeyFields() --> set

        Returns the list of fields in store that are not pkeys.
        """
        return set(self._store.keys()) - self.pkeys

    def _invalidFields(self,fields):
        """ _invalidFields(fields) --> set

        field in fields are returned if they are not present
        in self.fields.
        """

        if type(fields) is not set:
            fields=set(fields)
        return fields - self.fields

    def _fieldsMissing(self):
        """ _fieldsMissing() --> set

        Returns the list of table fields not valu'ed yet.
        """
        return self.fields - set(self._store.keys())

    def __str__(self):
        """ __str__() --> string 
        """
        u= INF_COL("Unset")
        res= ("  Table |%s|, oid %s:\n\n"\
             % (OKI_COL(self.table_name), 
                 self.oid and OKI_COL(str(self.oid)) 
                 or u))
        sget= self._store.get
        for k in self.fields:
            val= sget(k,u)
            sep= k in self._moved and '*' or '~'
            res+= ("  %12s %s %s\n" % (k,sep,str(val)))
        return res

#
##
### Sql Formatting
##
#

    def _sqlWhereOid(self):
        """ _sqlWhereOid() --> string

        Returns 'WHERE oid=self.oid' if self.oid. Crash otherwise.
        """
        if not self.oid:
            self.log_int_raise("Mth: NULL oid. Shouldn't be here.") 
        return (" oid=%d " % self.oid)

    def _sqlUpdate(self, d, filter= None):
        res= []
        s= self._store
        for k in d:
            v= s[k]
            if v== SQL_DEFAULT:
                res.append('%s= DEFAULT' % k)
            else:
                res.append('%s= %%(%s)s' % (k, k))

        if filter is None:
            filter= self._sqlWhereOid()

        qry= ("UPDATE %s SET %s WHERE %s" % 
                (self.table_name, ', '.join(res), filter))
        return self.mogrify(qry, s)

    def _sqlInsert(self, d):
        what= ', '.join(d)
        res= []
        s= self._store
        for k in d:
            v= s[k]
            if v== SQL_DEFAULT:
                res.append('DEFAULT')
            else:
                res.append('%%(%s)s' % k)

        qry= ("INSERT INTO %s (%s) VALUES (%s)" % 
                (self.table_name, what, ', '.join(res)))
        return self.mogrify(qry, s)

    def _sqlDelete(self, filter= None):
        if filter is None:
            filter= self._sqlWhereOid()
        return ("DELETE FROM %s WHERE %s" % 
                (self.table_name, filter))

    def _sqlSelect(self, d, filter= None, tbl= None):
        if filter is None:
            filter= self._sqlWhereOid()
        tbl= tbl or self.table_name
        what= ['%s.%s' % (tbl, f) for f in d]

        return ("SELECT %s FROM %s WHERE %s" % 
                (', '.join(what), tbl, filter ))


    def _sqlBuildFilter(self, filter, prefix= False, tbl= None):
        """ _sqlBuildFilter(self, filter [,prefix, tbl]) --> str

        Build the filter as SQL string. filter could be
        a string, a dict or a list (in such case use internal store).
        If prefix, WHERE is added (unless it's present yet).
        With tbl you can specify where apply the filter.
        """

        t= lambda tp: isinstance(filter, tp)
        tbl= tbl or self.table_name

        if t(str):
            s= filter

        elif t(dict):
            res= []
            for k, v in filter.iteritems():
                if v== SQL_DEFAULT:
                    res.append('%s.%s= DEFAULT' % (tbl, k))
                else:
                    res.append('%s.%s= %%(%s)s' % (tbl, k, k))
            s= self.mogrify(' AND '.join(res), filter)

        elif t(set) or t(frozenset) or t(list):
            res= []
            for k in filter:
                v= self._store[k]
                if v== SQL_DEFAULT:
                    res.append('%s.%s= DEFAULT' % (tbl, k))
                else:
                    res.append('%s.%s= %%(%s)s' % (tbl, k, k))
            s=self._sqlMogrify(' AND '.join(res))

        else:
            self.log_int_raise(\
                    "Invalid filter type: %s.",\
                    ERR_COL(str(filter)))

        # Prevent Double WHERE
        if not prefix or 'WHERE' in s:
            return s

        return " WHERE %s " % s 

    def _sqlMogrify(self, s, d= None):
        """ A wrapper to dbda.mogrify that take
        self._store as default dict value.
        """
        return self.mogrify(s, d or self._store)


    def isPureRelation(self,params=None):
        """ isPureRelation --> None

        The MoThEr Core, the Holy Hearth.
        """
        if params is None:
            return "I s P u R e R e L a T i O n !"
        else:
            return "Mi Sembva Di Impazzive."

    def get_triggers(self, flag, when):
        """ get_triggers(flag, when) -> list(functions)

        Returns triggers.
        """

        if not self._triggers_are_initialized():
            self.log_int_raise("Trigger are not initialized.")

        t= self.triggers_map
        try:
            d= t[flag]
        except:
            self.log_int_raise("get_trigger(): invalid flag %s.", ERR_COL(flag))

        try:
            l= d[when]
        except:
            self.log_int_raise("get_trigger(): invalid time %s.", ERR_COL(when))

        return l

    def get_flag_triggers(self, flag):
        """ get_flag_triggers(flag) -> list(functions)

        Returns triggers.
        """
        return  self.get_triggers(flag, MO_BEFORE) + \
                self.get_triggers(flag, MO_AFTER)

    def has_trigger(self, flag):
        """ has_trigger(flag) -> bool

        Is there a trigger for the flag `flag`?
        """
        if not self._triggers_are_initialized():
            return False
        return len(self.get_flag_triggers(flag)) > 0

    def add_trigger(self, flag, when, trigger):
        """ add_trigger(flag, when, f) --> None

        flag= MO_LOAD, MO_SAVE, MO_DEL, MO_UP
        when= MO_AFTER, MO_BEFORE

        f is the trigger: a python function
        """

        if not self._triggers_are_initialized():
            self.initTriggers()

        self.triggers_map[flag][when].append(trigger)

    def trigger(self, flag, when, *args):
        """ trigger(flag, when, *args) --> None

        flag= MO_LOAD, MO_SAVE, MO_DEL, MO_UP
        when= MO_AFTER, MO_BEFORE

        execute the tasked triggers for action flag and when
        """

        if not self._triggers_are_initialized():
            self.log_int_raise("Triggers are not initialized.")

        l= self.get_triggers(flag, when)
        for f in l:
            f(*args)
            self.log_info("Trigger %s (%s-%s) Fired!", 
                    OKI_COL(f.func_name), _MO_WHEN_STRS[when], _MO_FLAGS_STRS[flag])

    def setField(self, field_name, field_value):
        """ setField(field_name, field_value) --> None

        set field. If field is an invalid field, CRASH.
        """

        # Pkeys cannot be changed! It's a non-sense.
        # If you want insert a new record, create new MoThEr 
        # If you want to load a different record, use refresh()
        if field_name in self.pkeys:
            self.log_int_raise(
                " Safe behaviour: you can not change Pkeys Value. "\
                "Want to load a new record? Use refresh() "\
                "Want to update a Pkey? Create new and delete old."
                )

        if field_name not in self.fields:
            self.log_int_raise("Mth: cannot setField(): field "\
                    "%s is invalid.",ERR_COL(field_name))

        # Ok, new value is good
        self._store[field_name]=field_value
        # Remember that a field is changed.
        self._moved.add(field_name)

    def setFields(self,fdict,safe_mode=True):
        """ setFields(fdict [,safe_mode= True]) --> None

        Update the store with fdict. 
        If `safe_mode`, it will be forbidden to change 
        primary keys value. Really, use safe_mode iff
        you know exactly what are you doing.

        """
        # Is fdict a valid dict?
        new_fields=set(fdict.keys())
        onzo=self._invalidFields(new_fields)
        if onzo:
            self.log_int_raise("Mth: cannot setFields(): "\
                    "invalid fields %s.", ERR_COL(str(onzo)))

        # Cannot change pkeys
        if safe_mode:
            sp=self.pkeys
            if sp & new_fields:
                self.log_int_raise(
                    "For your safety, you can not change Pkeys Value. "
                    "Want to load a new record? Use refresh(). "
                    "Want to update a Pkey? Create new and delete old. "
                    "If you know what you do, use safe_mode=False. "
                    )

        # Ok, new values are good.
        self._store.update(fdict)
        # Remember fields changed.
        self._moved |= new_fields

    def getFields(self, fields= None, autoload= False):
        """ getFields([fields=None,autoload=False]) --> dict

        Returns stored values.
        If fields is not None, returns values for fields specified.
        If a field is invalid, CRASH.
        If a field is unvalued:
            if autload: load
            if not autoload: CRASH
        """

        if fields is None:
            return self._store.copy()

        onzo=self._invalidFields(fields)
        if onzo:
            self.log_int_raise("Mth: cannot setFields(): "\
                    "invalid fields %s.", ERR_COL(str(onzo)))
        newd={}
        if autoload:
            skeys=set(self._store.keys())
            nkeys=set(fields)
            lkeys=nkeys-skeys
            if len(lkeys):
                self.log_insane("Autoloading fields %s...", INF_COL(lkeys))
                self.load(lkeys)
            
        for f in fields:
            try:
                val=self._store[f]
                newd[f]=val
            except:
                self.log_int_raise(\
                    "Mth: field %s is not valued.",\
                        ERR_COL(f))
        return newd

    def _insert(self):

        # Is record stored yet?
        if self.oid:
            self.log_int_raise("Mth: cannot insert(): record is attached, "\
                "OID=%s. Want to use update()?",ERR_COL(self.oid))

        s= self._store
        # Debugging control and informations.
        if len(s)<len(self.fields):
            self.log_insane( 
                    "Mth: key values not completely specified:"
                    " assuming DEFAULT value.")

        # Useful vars
        sk=s.keys()

        # Make the query ...
        qry= self._sqlInsert(sk)

        #  execute it ...
        self.oc_query(qry)
        # and get last oid
        self.oid=self.lastrowid()

        # Useful debugging infos...
        self.log_insane(
                "Mth: %s saved with oid %s.",
                OKI_COL(str(s)),OKI_COL(str(self.oid)))

        # Load primary keys if not specified  
        # on INSERT
        ing= self._unvaluedPKeys()
        if ing:
            self.log_insane("Loading missing pkeys %s",
                    INF_COL(ing))
            self._load(ing)

        else:
            self.log_insane("No pkey has to be loaded.")

        # Nothing has to be updated now.
        self._moved.clear()

    def _update(self):

        # Really need to update?
        if not self._moved:
            self.log_int_raise("Mth: nothing to update.")

        if self.oid:
            filter= None
        else:
            ange= self._unvaluedPKeys()
            if ange:
                self.log_raise("Refusing to UPDATE record "\
                                "without primary key (missing %s).",
                                ERR_COL(list(ange)), DeleteError)
            params= (set(self._store) - set(self._moved)) 
            filter= self._sqlBuildFilter(params)

        # So, we got OID ;)
        # Make the query ...
        qry= self._sqlUpdate(self._moved, filter= filter)

        # else execute it
        try:
            self.oc_query(qry)
        except QueryError,ange:
            self.log_raise(str(ange), UpdateError)

        # Debugging infos...
        self.log_insane( "Mth: fields %s moved ;)",
                OKI_COL(self._moved))

        # Nothing has to be updated now.
        self._moved.clear()

    def _delete(self):

        if self.oid:
            filter= None
        # The values in store act as filters
        # when deleting
        else: 
            ange= self._unvaluedPKeys()
            if ange:
                self.log_raise("Refusing to DELETE record "\
                                "without primary key (missing %s).",
                                ERR_COL(list(ange)), DeleteError)
            # If oid is not present, load it.
            # If cannot do that, CRASH!
            filter= self._sqlBuildFilter(self._store)

        # Make the query ...
        qry= self._sqlDelete(filter= filter)

        # and execute it
        try:
            self.oc_query(qry)
        except QueryError, ange:
            self.log_raise(str(ange), DeleteError)

        # Debugging infos...
        self.log_insane("Removed record with OID %s", OKI_COL(self.oid))

        # Delete all informations...  
        self._moved.clear()
        self._store.clear()
        self.oid= None

    def _load(self, fields= None, safe_mode= True):

        # Which fields we need?
        mf= fields and set(fields) or set(self._fieldsMissing())

        if not mf <= self.fields:
            self.log_int_raise("Invalid Fields %s for table %s.",
                    ERR_COL(list(mf - self.fields)), self.table_name)

        if not len(mf):
            if self.oid:
                self.log_info("All fields loaded for this record.")
                return self._store.copy()
            else:
                self.log_int_raise("No fields specified for SELECT: table %s",
                        ERR_COL(self.table_name))

        if not self.oid:
            mf.add('oid')
            if safe_mode:
                ange= self._unvaluedPKeys()
                if ange:
                    self.log_raise("Refusing to LOAD record "\
                                "without primary key (missing %s). "
                                "Use ULOAD if you want to surf",
                                ERR_COL(list(ange)), SelectError)
            filter= self._sqlBuildFilter(self._store)
        else:
            filter= self._sqlWhereOid()

        # Make the query ...
        qry= self._sqlSelect(mf, filter= filter)

        # and execute it
        try:
            res= self.or_query(qry)
        except Exception, ange:
            self.log_raise("Unable to SELECT: %s", str(ange), SelectError)

        self.oid= self.oid or res.pop('oid')
        self.log_insane( "Loaded fields for OID=%s: %s",
                    OKI_COL(self.oid),OKI_COL(str(res)))

        # Update values on store
        self._store.update(res)

        # Forgot old to-update fields  
        self._moved-=mf

        # Useful return, that's all
        return res

    def _triggered_action(self, flag, *args):

        f= self._trigger_actions(self, flag)

        if not self._triggers_are_initialized() or \
                not self.get_flag_triggers(flag):

            res= f(*args)

        else:
            self.beginTrans()
            try:
                self.log_insane(
                        "Launching %s triggers...",
                        OKI_COL('before'))
                self.trigger(flag, MO_BEFORE)

                self.log_insane( 
                        "Executing %s action...",
                        OKI_COL(_MO_FLAGS_STRS[flag]))

                res= f(*args)

                self.log_insane( 
                        "Launching %s triggers...",
                        OKI_COL('after'))

                self.trigger(flag, MO_AFTER)

            except Exception, s:
                self.rollback()
                self.log_int_raise(str(s))

            self.commit()

        self.log_info("Action %s.", OKI_COL("completed"))
        return res

    def delete(self):
        """ delete() --> None

        Really do you want to kill me?
        """
        self._triggered_action(MO_DEL)
    
    def load(self, fields= None, safe_mode= True):
        """ load([fields= None, safe_mode= True]) --> dict

        Retrieve record values from DB. 
        if fields is not None, only fields specified will
        be loaded.
        If safe_mode is False, try to load the record also if
        primary keys was not given.
        """
        self._triggered_action(MO_LOAD,fields,safe_mode)

    def insert(self):
        """ insert() --> None

        Insert to db the values stored in store.
        For field with no value, SQL_DEFAULT is assumed.
        """
        self._triggered_action(MO_SAVE)
        
    def update(self, updict= None):
        """ update([updict= None]) --> None

        Make a UPDATE statement for modified fields and/or updict.
        If updict is None, only fields setted with setField(s) are
        moved. Otherwise, also fields specified in the dict updict 
        are considered.
        """
        if updict:
            self.setFields(updict)

        self._triggered_action(MO_UP)

    def uload(self, fields= None):
        """ uload([fields= None]) --> dict

        This is a call to load() with safe_mode= False
        """

        return\
            self.load(fields,safe_mode=False)

    def refresh(self,d,flag=MO_NOA):
        """ refresh(d[,flag=MO_NOA]) --> None

        You can secify here a new set of pkeys and
        have a new, fresh mother
        """
        # Detach record from db
        self.oid=None
        # init store.
        self._initStore(d,flag)


class MotherFly(DbMother):
    def __init__(self, tbl):
        self.table_name= tbl

    def __call__(self, d= {}, flag= MO_NOA, oid= None, session= None):
        DbMother.__init__(self, d, flag, oid, session)
        return self


def getMotherBuilder(tbl):
    """ getMotherBuilder(tbl_name) --> class

    Returns a `MoThEr Builder On The Fly` Class.
    You can init the class with a simple call:

        momma= getMotherBuilder(table)
        momma(mydict, myflag)
    """
    momma= MotherFly(tbl)
    return momma


def getMotherObj(tbl, d={}, flag= MO_NOA, oid= None, session= None):
    """ getMotherObj(tbl_name[, diz= {}, flag= MO_NOA, oid= None, session= None]) --> obj

    Initialize and return a `MoThEr On The Fly` Instance.
    """
    momma= getMotherBuilder(tbl)
    momma(d, flag, oid, session)
    return momma

#
##
### The MoThEr Manager
##
#
#
##
###
#
# You can use the manager plugin to handle a set
# of child records, and to handle relations in a 
# magic way
# To use it, specify the dependencies:
#
#   class Foo(DbMother,MotherManager):
#       def __init__(self,.....):
#           DbMother.__init__(self,...
#
# but don't init the manager.
# You will have a set of useful functions.
# If you want more magic, you can run:
#
#   initChildManager(children_list)
#   initRelationManager(children_list)
#   initListManager(children_list)
#
###
##
#



class MotherManager:
    """ Mother Manager-Plugin. Need Mother."""

#
##
### Internal use.
##
#

    def _sqlJoin(self, j_tbl, r_tbl, j_key, r_key):
        """ _sqlJoin(self, j_tbl, r_tbl, j_key, r_key) --> str\n 

        Returns the join query.
        """

        s= " JOIN %s on " % r_tbl
        s+="AND".join(\
            [" %s.%s = %s.%s " % (r_tbl,v,j_tbl,k)
                for k,v in j_key.iteritems()])
        s+=" WHERE "
        s+="AND".join(\
            ["%s.%s = %%(%s)s " % (r_tbl,v,k)\
                for k,v in r_key.iteritems()])

        return s

    def _sqlJoinParent(self, builder, j_builder, outer=False):
        """ _sqlJoinParent(self, j_builder, builder [, outer= False]) --> (string, string)

        Returns a filter and the table where filter has to be applied.
        """
        tcd, trd= MotherMappedDeps()

        tbl= builder.table_name
        j_tbl= j_builder.table_name

        if j_tbl not in tcd[tbl].keys():

            rel= frozenset([tbl,j_tbl])

            try:
                j_tbl= trd[rel]
            except:
                self.log_int_raise('_sqlJoinParent: invalid join'\
                        ' table %s!' % ERR_COL(j_tbl))

            self.log_insane( 
                    "Joining relations %s", 
                    INF_COL(j_tbl))

        else:
            self.log_insane( 
                    "Joining child %s", 
                    INF_COL(j_tbl))

        m_d= self.getChildDeps(j_tbl, builder)

        if outer:
            s=" LEFT OUTER JOIN %s on " % j_tbl
        else:
            s=" JOIN %s on " % j_tbl

        s+="AND".join(\
            [" %(tbl)s.%(k)s = %(j_tbl)s.%(v)s " % locals()
            for k,v in m_d.iteritems()])

        return s, j_tbl

    def _getRelationObject(self, builders):
        """ _getRelationObject(builders): --> object

        object is a MotherOnTheFly object.
        builders is a list of table names.
        """
        
        rel_tbl_name= self._searchRelation(builders+ [self.table_name])

        # Return a builder with MotherOnTheFly.
        rel_obj= getMotherObj(rel_tbl_name, session= self.session)
        return rel_obj

    def getChildDeps(self, tbl_name, father= None):
        """ getChildDeps(self, tbl_name [, father= None]) --> dict

        Wich fields in TABLE tbl_name will be inherited from father?
        Returned dict is the mapping_dict.
        If father is None, self is assumed.
        """

        father=father or self
        tbl=father.table_name

        # Load father children.....
        dd= MotherMappedChildren()
        d= dd[tbl]

        # Get the dependencies for child tbl_name
        try:
            dd= d[tbl_name]
        except:
            self.log_int_raise(\
                "Child %s not specified for table %s.",\
                    ERR_COL(tbl_name),ERR_COL(tbl))
        return dd

    def _getDepDict(self, mapping_dict, father= None):
        """_getDepDict(mapping_dict [,father= None]): ---> dict

        Given the dependencies mapping_dict, a new dict is created,
        filled and returned with father values.
        If father is None, self is assumed.
        """

        father= father or self
        new_d={}

        self.log_insane( 
                "Exporting values %s ",
                str(mapping_dict))
                        
        fkeys=set(mapping_dict.keys())
        try:
            d= father.getFields(fkeys)
        except Exception, s:
            self.log_int_raise("Cannot inherit needed values from father: %s.",
                    ERR_COL(list(fkeys)))
            
        for mapkey in mapping_dict:
            mapped_field= mapping_dict[mapkey]
            new_d[mapped_field]= d[mapkey]

        return new_d

    def _fillChildDict(self, child_obj, father= None):
        """_fillChildDict(child_obj [,father= None]) --> None
        Given a child object, his store is updatetd with
        father dependencies. If father is None, father= self.
        """
        
        father= father or self
        ct= child_obj.table_name
        
        # Ge dependencies ...
        d= self.getChildDeps(ct,father)
        # Create new dict ...
        new_d= self._getDepDict(d,father)
        # Init child with new_dict
        child_obj.setFields(new_d,safe_mode=False)

        self.log_insane( 
                "Child object(%s) updated with values %s.",\
                OKI_COL(ct),OKI_COL(new_d))

    def _workChild(self, child_builder, child_dict, flag):
    
        # Build obj...
        new_obj= child_builder(child_dict, session= self.session)
            # Inherit father fields...
        self._fillChildDict(new_obj)
                # Get properly function...
        f= new_obj._flag_actions(new_obj, flag)
                    # Get some debugging vars
        tbl= new_obj.table_name
        sd= str(new_obj.getFields())
        s= _MO_FLAGS_STRS[flag]
                # Execute function....
        f()
            # Debug
        self.log_insane( 
                "%s child %s, values %s",
                s,tbl,sd)

        return new_obj
        # Return ;)

    def _searchRelation(self, tbls):

        fz= frozenset(tbls)

        trd= MotherMappedRels()
        try:
            r_tbl= trd[fz]
            self.log_insane(
                    "Using relation %s to join "
                    "tables %s", r_tbl,tbls)
            return r_tbl

        except:
            pass

        self.log_info(\
            "Warning: direct relation for set %s "
            "is not defined: searching subset...",
            INF_COL(fz))

        rels= [trd[k] for k in trd.keys() if fz< k]

        if len(rels)<>1:
            self.log_int_raise(
                "In getRelation: can not get the "  
                "relation table. There are %s possible "     
                "choices: %s .", ERR_COL(len(rels)), ERR_COL(
                ",".join(rels)))

        return rels[0]

       
#
##
### Public
##
#

    def getChild(self, c_builder, c_dict):
        """ getChild(c_builder, c_dict): --> MotherObj
        """
        return self._workChild( 
                                c_builder,
                                c_dict,
                                MO_LOAD,
                              )
                                    

    def insertChild(self, c_builder, c_dict):
        """ insertChild(child_builder, c_dict) -> MotherObj
        insert new child object and return it.
        """
        return self._workChild( 
                                c_builder,
                                c_dict,
                                MO_SAVE,
                              )

    def updateChild(self,c_builder,c_dict):
        """ updateChild(c_builder, c_dict): --> MotherObj
        """
        return self._workChild( 
                                c_builder,
                                c_dict,
                                MO_UP,
                              )
        
    def deleteChild(self,c_builder,c_dict):
        """ deleteChild(c_builder, c_dict): --> MotherObj
        """
        return self._workChild( 
                                c_builder,
                                c_dict,
                                MO_DEL,
                              )

    def deleteChildren(self, c_builder, filter= None):
        """ deleteChildren(c_builder [,filter=None]) --> None
        
        Delete a set of children, filtered with filter.
        If filter is None, delete all children.
        c_builder is a Mother class for the table where records
        wil be deleted.
        """
        child_table= c_builder.table_name
        d= self.getChildDeps(child_table)
        d= self._getDepDict(d)

        qry_filter= self._sqlBuildFilter(d, tbl=child_table)
        if filter is not None:
            qry_filter+= " AND %s" % \
                    self._sqlBuildFilter(filter, tbl=child_table)

        MotherBox(c_builder, qry_filter, MO_DEL)

    def updateChildren(self, c_builder, d_up, filter= None):
        """ updateChildren(self, c_builder, d_up [, filter= None]) --> None

        Update children with dict d_up.
        If filter, update only filtered children.
        c_builder is a Mother class for the table.
        """
        child_table= c_builder.table_name
        d= self.getChildDeps(child_table)
        d= self._getDepDict(d)

        # Build filter
        qry_filter= self._sqlBuildFilter(d, tbl=child_table)
        if filter is not None:
            qry_filter+= " AND %s" % \
                    self._sqlBuildFilter(filter, tbl=child_table)

        MotherBox(c_builder, flag=MO_UP, fields= d_up, 
                filter= qry_filter, session= self.session) 

    def getChildren(self, c_builder, fields= None, filter= None,\
                    builder_join= None, filter_join= None,
                    outer= False, order= None):
        " getChildren(c_builder [,fields= None, filter= None, "\
        " builder_join= None, filter_join= None, outer= False, order= None])-->box\n" \
        """
        fields = list of fields names
        filter = string or dictionary
        builder_join = builder class which will be joined
        filter_join = string or dictionary about join table fields
        outer = boolean, will be a left outer join
        order = list(fields)
        """

        child_table= c_builder.table_name
        # Get dependencies dict...
        d= self.getChildDeps(child_table)
        d= self._getDepDict(d)

        # If a join filter is specified, we need
        # to build the JOIN STATEMENT before the WHERE
        qry_filter, joining_table= builder_join and \
            self._sqlJoinParent(c_builder, builder_join, outer) or \
                ("", "")

        # Build SQL filter string
        qry_filter+= " WHERE "

        # now filter father dependencies
        qry_filter+= self._sqlBuildFilter(d, tbl= child_table)

        # Filter join condition
        # Note that the filter is used on the joined table
        # that is returned by _sqlJoinParent
        if builder_join and filter_join:
            qry_filter+=" AND %s " %\
                self._sqlBuildFilter(\
                filter_join, tbl= joining_table)

        # if filter is specified, use it ...
        # the filter `filter` is used on Child Table.
        if filter is not None:
            qry_filter+=" AND  %s " %\
                self._sqlBuildFilter(filter,tbl=child_table)

        # Let's use a MotherBox :)
        # If join use DISTINCT
        distinct= (builder_join is not None)
        mb= MotherBox(c_builder, filter= qry_filter, flag= MO_LOAD,
                        fields= fields, order= order, distinct= distinct, 
                        session= self.session)
        # Oh yes, return ;)
        return mb

    def assignRelation_nt(self, builder_list, params= {}):
        """ """

        # Handle given objects ...
        builders= []
        fathers= []
        self.log_insane("Creating auto-relation...")

        for builder, d, flag in builder_list:
            tbl= builder.table_name
            builders.append(tbl)
            flag_str= _MO_FLAGS_STRS[flag]
            self.log_insane(
                    "Working record in %s with "\
                    "values %s and flag=%s...",\
                    INF_COL(tbl),INF_COL(str(d)),flag_str)
            try:
                new_obj= builder(d, flag, session= self.session)
                fathers.append(new_obj)
            except Exception,s:
                es=("Error! %s initialization returned: %s."\
                              " Now Rollback!" % (flag_str,ERR_COL(str(s))))
                self.log_error(es)
                self.log_int_raise("Unable to assign child: %s",es)

        self.log_insane( 
                "%s child(ren) assigned, now work on "\
                "relation...", OKI_COL(len(builders)))

        # create auto relation ...
        # init child object 
            
        # self.table_name is added automatically
        r_obj = self._getRelationObject(builders)

        # Let child take his values from fathers...
        for father in [self]+fathers:
            self._fillChildDict(r_obj, father)
        r_obj.setFields(params, safe_mode=False)

        # store it to DB
        try:
            r_obj.insert()
        except Exception,s:
            self.log_error("Error! insert() returned: %s."\
                                " Now Rollback!",ERR_COL(str(s)))
            self.log_int_raise("Error adding auto-relation.") 

        # All was fine.
        self.log_insane( 
                "Relation %s auto-saved with "\
                "values %s.",r_obj.table_name,\
                OKI_COL(str(r_obj._store)))

        return tuple(fathers)

    def assignRelation(self, builder_list, params= {}):
        """ assignRelation(self, builder_list [,params= {}]) --> Tuple

        assign a Relation.
        builder_list is a list of tuples:

            [(MotherBuilder,dict,flag)...]

        If TABLE A and TABLE B are related with the relation TABLE
        R_AB, you can (pseudo-code):

            me= Mother(A)
            builder_list=[(B, dict, MO_SAVE)]
            me.assignRelation(builder_list)

        With this call, create a record in B and put in R_AB the 
        relation.
        If the relation accept other fields, specify them in params,
        which is a dict.
        """

        self.beginTrans()
        try:
            res=self.assignRelation_nt(builder_list,params)
        except Exception,s:
            self.rollback()
            raise Exception(s)

        self.commit()
        return res
            
    def dropRelation(self, builder_list, params= {}):
        """ dropRelation(self, builder_list [,params= {}]) --> list

        Drop relation between father in builder_list.
        builder_list is [(builder,dict,flag),...]
        """
        fathers=[]
        fathers_del=[]
        # Create the list of only builders
        for builder, d, flag in builder_list:
            new_obj= builder(d, session= self.session)
            fathers.append(new_obj)
            if flag==MO_DEL:
                fathers_del.append(new_obj)

        # create auto relation init child object
        t_names=[o.table_name for o in fathers]
        r_obj = self._getRelationObject(t_names)
        
        # Let child take his values from fathers...
        for father in [self]+fathers:
            self._fillChildDict(r_obj,father)

        r_obj.setFields(params)
        r_obj.delete()
        self.log_insane("Dropped relation %s",\
                OKI_COL(r_obj.table_name))
        # Destroy all fathers with flag MO_DEL
        for f in fathers_del:
            f.delete()
        return fathers
    
    def dropRelations(self, builder, filter= None, r_tbl= None, flag=MO_NOA):
        """ dropRelation(builder [,filter=None, r_tbl= None, flag= MO_NOA]):

        drop all related records.
        If r_tbl is specified, use that table to join records,
        otherwise, r_tbl is deduced.

        fIlter is.. a filter.
        """

        mbox=self.joinRelation(builder, filter= filter, r_tbl= r_tbl)

        self.log_insane( 
                "%s is dropping %s relation(s) ...",
                INF_COL(self.table_name), INF_COL(len(mbox)))

        ds= mbox.getRecords()
        for d in ds:
            self.dropRelation([(builder, d, flag)])

    def joinRelation(self, builder, fields= [], filter= None, r_tbl= None, order= None):
        """ joinRelation(self, builder, fields= [], filter= None, r_tbl= None, order= None) -> Box

        Get related data.
        Builder is the MotherBuilder for data to be searched.
        Filter could be a string or a dict: if given, query will
        be restricted. Note that filter will be applied to the
        relation table.
        r_tbl is the relation table used for the join. If not given,
        r_tbl is obtained automaically.
        if order is not None, the query will be ordered by fields contained in that list.
        """

        # Useful vars
        s_tbl=self.table_name
        b_tbl=builder.table_name

        # Try to get relation table....
        # If there is ambiguity, CRASH!
        if r_tbl is None:
            r_tbl= self._searchRelation([s_tbl, b_tbl])

        # Ok, the relation table is found
        # Getting the mapping child dicts
        s_d = self.getChildDeps(r_tbl)
        b_d = self.getChildDeps(r_tbl, builder)
        
        # If fields is not specified, fields for the
        # serached table are retrieved 
        if fields== []:
            tfd= MotherMappedFields()
            fields= tfd[b_tbl]
            del tfd

        self.log_insane("Joining %s and %s with relation %s "\
                        "and fields %s ...",s_tbl,b_tbl,\
                        INF_COL(r_tbl),INF_COL(str(fields)))
                        
        # Make the filter query ...
        qry_filter = self._sqlJoin(b_tbl, r_tbl, b_d, s_d)
        # If filter is given, apply it.
        if filter:
            qry_filter+=" AND %s " % self._sqlBuildFilter(filter, tbl= r_tbl)
        # Mogrify the obtained filter query ...
        qry_filter= self._sqlMogrify(qry_filter)

        # get a MotherBox
        mb= MotherBox(builder, qry_filter, MO_LOAD, fields, order, session= self.session)
        return mb

    def initChildManager(self, children):
        """initChildManager(self,children) --> None
        
        You can define a Mother class for each Child-TABLE.
        In fact, children is a list of Mother Builders.
        The following functions will be wrappered:
            
            insertChild(Foo,dict)
            deleteChildren(Foo, filter)
            updateChildren(Foo, dict, filter)
            getChildren(builder,fields,filter,builder_join,filter_join,outer,order)


        So, if FOO is a child, you can create the Mother Class
        Foo and have the methods:

            insertFoo(dict)
            deleteMultipleFoo(filter)
            updateMultipleFoo(dict, filter)
            getMultipleFoo(fields, filter, builder_join, filter_join, outer, order)

        * NOTE: methods already defined are not overwritten.
        """

        insertchild= lambda builder:(                   \
                        lambda idict:                   \
                            self.insertChild(builder, idict))

        deletechildren= lambda builder:(                \
                            lambda filter=None:         \
                                self.deleteChildren(builder, filter))

        getchildren= lambda builder:(                                                   \
                        lambda fields=None,filter=None,                                 \
                            builder_join=None,filter_join=None,outer=None,order=None:   \
                                self.getChildren(                                       \
                                            builder,                                    \
                                            fields,                                     \
                                            filter,                                     \
                                            builder_join,                               \
                                            filter_join,                                \
                                            outer,                                      \
                                            order))

        updatechildren= lambda child:(                  \
                            lambda d_up, filter= None:  \
                                self.updateChildren(child, d_up, filter))

        camel= DbMother.MotherNaming
        for c in children:
            ctbl=c.table_name

            attr_name= camel(ctbl,'insert')
            if not hasattr(self, attr_name): 

                attr= insertchild(c)
                attr.__doc__=\
                        " %s(dict) --> child\n\n"                                       \
                        "insert a child. dict contains the fields values.\n"            \
                        "Don't specify the foreign key: it's assigned automatically.\n" \
                            % attr_name

                setattr(self, attr_name, attr)

            attr_name= camel(ctbl, 'deleteMultiple')
            if not hasattr(self, attr_name): 
                
                attr= deletechildren(c)
                attr.__doc__=\
                        " %s(filter= None) --> None\n\n"                                \
                        "delete children. if filter is None, delete ALL children."      \
                            % attr_name

                setattr(self, attr_name, attr)

            attr_name= camel(ctbl, 'getMultiple')
            if not hasattr(self, attr_name): 
                
                attr= getchildren(c)
                attr.__doc__=\
                        " %s([fields= None, filter= None, builder_join= None, "                 \
                        "filter_join, outer= None, order= None) --> MotherBox\n\n"              \
                        "retrieve a set of children. filter is used for the children table.\n"  \
                        "builder_join and filter_join are used to do advanced filtering.\n"     \
                        "If outer is True, an outer join is performed.\n"                       \
                        "order is a list of fields names." % attr_name

                setattr(self, attr_name, attr)

            attr_name= camel(ctbl, 'updateMultiple')
            if not hasattr(self, attr_name): 
                
                attr= updatechildren(c)
                attr.__doc__=\
                        """ %s(d_up, filter= None) --> None\n\n\
                        update children. the dict d_up contains the fields (and their values)
                        to be updated. The filter is a usual filter.
                        """ % attr_name

                setattr(self, attr_name, attr)

    def initRelationManager(self,children):
        """initRelationManager(children) --> None
        
        This is a magic function.
        It's a wrapper for the methods:

            - assigRelation
            - joinRelation
            - dropRelation
            - dropRelations

        The wrappers will be:

            - assigFoo
            - joinFoo
            - dropFoo
            - dropMultipleFoo

        NOTE: methods already defined are not overwritten.
        """

        d={
            self.assignRelation:"assign",
            self.dropRelation:"drop"
        }

        magic = lambda handler,child:(\
                lambda d,flag=MO_NOA,params={}:\
                    handler([(child,d,flag)],params)[0])

        b={
            self.joinRelation:'join',
            }
            
        logic = lambda handler,builder:(\
                lambda fields=[],\
                       filter=None,\
                       r_tbl=None,\
                       order=None:\
                        handler(builder,fields,\
                                 filter,r_tbl,order))

        g={
            self.dropRelations:'dropMultiple',
        }

        gigic = lambda handler,child:(\
                lambda filter=None,\
                       r_tbl=None,\
                       flag=MO_NOA:
                        handler(child,filter,r_tbl,flag)) 

        camel= DbMother.MotherNaming
        for child in children:
            c=child.table_name
            for f,pref in d.iteritems():
                attr=magic(f,child)
                attr_name=camel(c,pref)
                # Don't let initRelationManager overwrite
                # user methods
                if hasattr(self,attr_name):
                    continue
                attr.__doc__=\
                    """ %s(dict[,flag=MO_NOA,params={}]) --> obj

                    Handle the Mother object created with dict and flag,
                    but after insert one record on the relation table that
                    join %s and %s.
                    Return the handled Mother Object."
                    """ % (attr_name,self.table_name,c)
                setattr(self,attr_name,attr)
            
            for f,pref in b.iteritems():
                attr=logic(f,child)
                attr_name=camel(c,pref)
                # Don't let initRelationManager overwrite
                # user methods
                if hasattr(self,attr_name):
                    continue
                attr.__doc__=\
                    """ %s([,fields,filter,r_tbl,order]) --> MotherBox

                    Do a join with records in %s.
                    If filter is specified, apply it.
                    If r_tbl is specified, use that as relation table to make the join.
                    If fields is specified, load only these fields from the joined table.
                    Return a MotherBox object.
                    """  % (attr_name,c)
                setattr(self,attr_name,attr)

            for f,pref in g.iteritems():
                attr=gigic(f,child)
                attr_name=camel(c,pref)
                # Don't let initRelationManager overwrite
                # user methods
                if hasattr(self,attr_name):
                    continue
                attr.__doc__=\
                    """ %s(filter,r_tbl,flag) --> None

                    Drop a set of relations.
                    if flag is specified, do the relative action
                    on related records.
                    if r_tbl is specified, use that table to make the join.
                    """  % (attr_name)
                setattr(self,attr_name,attr)

#
##
### The MotherBox
##
#

class MotherBox(Speaker, Psygres):
    def __init__(self, builder, filter= None, flag= MO_NOA, 
                    fields= None, order= None, session= None, distinct= False):

        if session:
            session._import(self)

        self.session= session
        self.builder= builder
        self.momma= builder()
    
        if filter:
            filter= self.momma._sqlBuildFilter(filter, True)
        else:
            filter= ''

        self._store= []

        if flag==MO_LOAD:
            self.loadBox(filter, fields, order, distinct)

        elif flag==MO_DEL:
            self.deleteBox(filter)

        elif flag==MO_UP:
            if not fields:
                self.log_int_raise(
                        "No fields specified to update %s.", 
                        ERR_COL(self.builder.table_name))
            self.updateBox(filter, fields)

        elif flag== MO_NOA:
            return

        else:
            self.log_int_raise("Invalid Flag %s for MotherBox", ERR_COL(flag))

        del self.momma


    def __len__(self):
        return len(self._store)

    def _retrieve_oided_mothers(self, filter):

        where= filter or ''
        tbl= self.momma.table_name
        qry= "SELECT %(tbl)s.oid from %(tbl)s %(where)s" % locals()

        res= self.mr_query(qry)
        session= self.session
        mothers= [self.builder(oid= d['oid'], session= session) for d in res]

    def _trigger_oided_mothers(self, mothers, flag, when):
        for m in mothers:
            m.trigger(flag, when)

    def _loadBox(self, filter, fields, order, distinct):

        tbl= self.builder.table_name

        self.log_insane(
                "Mbox: loading with filter=%s",\
                INF_COL(filter))

        distinct= distinct or ''

        sel="SELECT %(distinct)s %(tbl)s.oid, %(what)s FROM %(tbl)s %(filter)s "

        if filter:
            filter= self.momma._sqlBuildFilter(filter, prefix= True)
        else:
            filter= ''

        if fields is None:
            what= "%s.*" % tbl
        else:
            ifields= set(fields) - self.momma.fields
            if len(ifields):
                self.log_int_raise("Inavlid Fields for table %s: %s",
                    tbl, ERR_COL(list(ifields)))
            what= ",".join(['%s.%s ' % (tbl, f) for f in fields])

        qry= sel % locals()

        if order:
            qry+= " ORDER BY %s" % ','.join(order)

        res= self.mr_query(qry)
        self._store= [ (d.pop('oid'), d) for d in res ]

        self.log_debug("Loaded %s records on %s",\
                OKI_COL(len(res)), tbl)

        self.log_info("Action %s.", OKI_COL("completed"))


    def loadBox(self, filter, fields, order, distinct):

        if not self.momma.has_trigger(MO_LOAD):
            self._loadBox(filter, fields, order, distinct)
            return 
            
        self.log_info("Launching Triggers on Box (select) operations...")
        ms= self._retrieve_oided_mothers(filter)
        self._trigger_mothers(ms, MO_LOAD, MO_BEFORE)
        self._loadBox(filter, fields, order, distinct)
        self._trigger_mothers(ms, MO_LOAD, MO_AFTER)

    def updateBox(self, d, filter= None):

        if not self.momma.has_trigger(MO_UP):
            self._updateBox(d, filter)
            return

        self.log_info("Launching Triggers on Box (update) operation...")
        ms= self._retrieve_oided_mothers(filter)
        self._trigger_mothers(ms, MO_UP, MO_BEFORE)
        self._updateBox(d, filter)
        self._trigger_mothers(ms, MO_UP, MO_AFTER)

    def _updateBox(self, filter, d):
        """ updateBox(self, d, filter=None): --> None

        Update a set of records.
        """

        table= self.builder.table_name

        if filter:
            ftr= self.momma._sqlBuildFilter(filter, prefix= True)
        else:
            ftr= ''

        res= []
        for k, v in d.iteritems():
            if v== SQL_DEFAULT:
                res.append('%s= DEFAULT' % k)
            else:
                res.append('%s= %%(%s)s' % (k, k))

        what= ', '.join(res)
        tbl= self.builder.table_name

        qry= 'UPDATE %(tbl)s SET %(what)s %(ftr)s' % locals()
        qry= self.mogrify(qry, d)

        self.oc_query(qry)
        self.log_insane("Box: Updatetd records.")

        self.log_info("Action %s.", OKI_COL("completed"))

    def deleteBox(self, filter=None):

        if not self.momma.has_trigger(MO_DEL):
            self._deleteBox(filter)
            return 
        
        self.log_info("Launching Triggers on Box (delete) operations...")
        ms= self._retrieve_oided_mothers(filter) 
        self._trigger_mothers(ms, MO_DEL, MO_BEFORE)
        self._deleteBox(filter)
        self._trigger_mothers(ms, MO_DEL, MO_AFTER)

    def _deleteBox(self, filter=None):
        """ deleteBox(self, filter= None, stored= False) --> None

        delete from DB.
        If stored is True, delete stored object.
        """

        self.log_insane("Mbox: deleting with filter=%s", INF_COL(filter))

        tbl= self.builder.table_name
        if filter:
            filter= self.momma._sqlBuildFilter(filter, prefix= True)
        else:
            filter= ''

        qry="DELETE FROM %(tbl)s %(filter)s" % locals()
        self.oc_query(qry)
        self.log_insane("Records deleted.")

        self.log_info("Action %s.", OKI_COL("completed"))

    def getRecords(self, flag_obj= False):
        """ getRecords(self, flag_obj= False) -> list(dict) or list(Mothers)

        The return value depends on flag_obj.
        """
        if flag_obj:
            b= self.builder
            session= self.session
            return [b(d, MO_NOA, oid, session) for oid, d in self._store]
        else:
            return [d for oid, d in self._store]

    def getFields(self, flag_obj= False):
        """ getRecords(self, flag_obj= False) -> list(dict) or list(Mothers)

        The return value depends on flag_obj.
        MotherBox.getFields is obsolete! Use MotherBox.getRecords!
        """
        self.log_warning(ERR_COL(
            "Obsolete Method MotherBox.getFields(). Use getRecords() instead! "
            "This method will be removed from Mother."))
        return self.getRecords(flag_obj)


